@@ -7,8 +7,9 @@ import shutil
 from collections import defaultdict
 from dataclasses import dataclass
 from datetime import datetime
+from json import JSONDecodeError
 from pathlib import Path
-from typing import Any, Dict, List, Optional, Set, TextIO, Type, Union
+from typing import Any, Dict, Iterable, List, Optional, Set, TextIO, Type, Union, cast
 
 import click
 import htmlmin
@@ -40,6 +41,29 @@ TYPE_NUMBER = "number"
 TYPE_OBJECT = "object"
 TYPE_STRING = "string"
 
+KW_REQUIRED = "required"
+KW_TITLE = "title"
+KW_CONTAINS = "contains"
+KW_ITEMS = "items"
+KW_UNIQUE_ITEMS = "uniqueItems"
+KW_MAX_ITEMS = "maxItems"
+KW_MIN_ITEMS = "minItems"
+KW_MAX_LENGTH = "maxLength"
+KW_MIN_LENGTH = "minLength"
+KW_PATTERN = "pattern"
+KW_CONST = "const"
+KW_ENUM = "enum"
+KW_ELSE = "else"
+KW_THEN = "then"
+KW_IF = "if"
+KW_NOT = "not"
+KW_ONE_OF = "oneOf"
+KW_ANY_OF = "anyOf"
+KW_ALL_OF = "allOf"
+KW_PROPERTIES = "properties"
+KW_PATTERN_PROPERTIES = "patternProperties"
+KW_ADDITIONAL_PROPERTIES = "additionalProperties"
+
 DESCRIPTION = "description"
 DEFAULT = "default"
 EXAMPLES = "examples"
@@ -76,6 +100,7 @@ class GenerationConfiguration:
     copy_js: bool = True
     link_to_reused_ref: bool = True
     recursive_detection_depth: int = 25
+    template_name: str = "js"
 
 
 class SchemaNode:
@@ -89,7 +114,9 @@ class SchemaNode:
         file: str,
         path_to_element: List[Union[str, int]],
         html_id: str,
-        is_property: bool = False,
+        ref_path="",
+        parent: "SchemaNode" = None,
+        parent_key: str = None,
         literal: Union[str, int, bool] = None,
         keywords: Dict[str, Union["SchemaNode", str, List[str]]] = None,
         array_items: List["SchemaNode"] = None,
@@ -102,7 +129,33 @@ class SchemaNode:
         :param file: Real path to the schema file
         :param path_to_element: Path from the root of the schema to the current element
         :param html_id: HTML ID for the current element. Used for anchor links.
-        :param is_property: If True, then this node is under "properties" and it should not be treated as a keyword
+        :param parent: The parent node of which the current node is an array item or keyword
+        :param parent_key: If the node is under a keyword of the parent node, that keyword
+            Example:
+            In the following context
+            {
+                "patternProperties": {
+                    ".*": {
+                        "type": string
+                    }
+                }
+            }
+
+            For the node
+            {
+                "type": string
+            }
+            The parent_key is ".*"
+
+            For the node
+            {
+                ".*": {
+                    "type": string
+                }
+            }
+            The parent key is "patternProperties"
+
+        :param ref_path: Path of a reference to this element, if any (usually "#/definitions/A name")
         :param literal: If the schema is neither a dict nor an array, it will be kept here
                         Useful for things like description, types, const, enum, etc.
         :param keywords: If the schema is a dict, this will be filled. Otherwise, this stays empty
@@ -115,17 +168,74 @@ class SchemaNode:
         self.depth = depth
         self.file = file
         self.path_to_element = path_to_element
+        self.html_id = html_id or "_".join(path_to_element) or "root"
+        self.parent = parent
+        self.parent_key = parent_key
+        self.ref_path = ref_path
         self.literal = literal
         self.keywords = keywords or {}
         self.array_items = array_items or []
         self.refers_to = refers_to
         self.is_displayed = is_displayed
-        self.html_id = html_id or "_".join(path_to_element) or "root"
-        self.is_property = is_property
 
     @property
-    def is_definition(self) -> bool:
-        return self.path_to_element and self.path_to_element[0] != "definitions"
+    def definition_name(self) -> str:
+        if self.ref_path:
+            return self.ref_path.split("/")[-1]
+        return ""
+
+    @property
+    def is_property(self) -> bool:
+        return bool(self.parent and self.parent.parent_key == KW_PROPERTIES)
+
+    @property
+    def is_pattern_property(self) -> bool:
+        return bool(self.parent and self.parent.parent_key == KW_PATTERN_PROPERTIES)
+
+    @property
+    def is_additional_properties(self) -> bool:
+        return self.parent_key == KW_ADDITIONAL_PROPERTIES
+
+    @property
+    def is_a_property_node(self) -> bool:
+        return self.is_property or self.is_pattern_property or self.is_additional_properties
+
+    @property
+    def is_additional_properties_schema(self) -> bool:
+        return self.is_additional_properties and self.literal is not True
+
+    @property
+    def iterate_properties(self) -> Iterable["SchemaNode"]:
+        if self.kw_properties:
+            yield from self.kw_properties.keywords.values()
+
+        if self.kw_pattern_properties:
+            yield from self.kw_pattern_properties.keywords.values()
+
+        if self.kw_additional_properties is not None and self.kw_additional_properties.literal is not False:
+            yield self.kw_additional_properties
+
+    @property
+    def required_properties(self) -> List[str]:
+        """The required properties for this node"""
+        if not self.parent:
+            return []
+
+        required_properties = self.parent.kw_required
+        if not required_properties:
+            return []
+
+        return [r.literal for r in required_properties.array_items]
+
+    @property
+    def is_required_property(self) -> bool:
+        return self.parent and self.property_name in self.parent.required_properties
+
+    @property
+    def path_to_property(self) -> str:
+        """Human-readable representation of the path from the root of the schema to this node"""
+        path_without_properties = [p for p in self.path_to_element if p not in [KW_PROPERTIES, KW_PATTERN_PROPERTIES]]
+        return " -> ".join([p if isinstance(p, str) else f"Item {p}" for p in path_without_properties])
 
     @property
     def flat_path(self) -> str:
@@ -137,7 +247,7 @@ class SchemaNode:
         possible_default = self.keywords.get(DEFAULT)
         if not possible_default:
             return None
-        if isinstance(possible_default, SchemaNode) and possible_default.is_property:
+        if isinstance(possible_default, SchemaNode) and possible_default.is_a_property_node:
             return None
         return possible_default
 
@@ -147,17 +257,135 @@ class SchemaNode:
         if not possible_examples:
             return []
 
-        if isinstance(possible_examples, SchemaNode) and possible_examples.is_property:
+        if isinstance(possible_examples, SchemaNode) and possible_examples.is_a_property_node:
             return []
 
         return possible_examples
 
     def get_keyword(self, keyword: str) -> Optional["SchemaNode"]:
+        """Get the value of a keyword if present and it is not a property (to avoid conflicts with properties being
+        named like a keyword, e.g. a property named "if")
+        """
         possible_keyword = self.keywords.get(keyword)
         if possible_keyword and isinstance(possible_keyword, SchemaNode) and not possible_keyword.is_property:
             return possible_keyword
+
         return None
 
+    @property
+    def kw_all_of(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ALL_OF)
+
+    @property
+    def kw_any_of(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ANY_OF)
+
+    @property
+    def kw_one_of(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ONE_OF)
+
+    @property
+    def kw_not(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_NOT)
+
+    @property
+    def has_conditional(self) -> bool:
+        return self.kw_if is not None and (self.kw_then is not None or self.kw_else is not None)
+
+    @property
+    def kw_if(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_IF)
+
+    @property
+    def kw_then(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_THEN)
+
+    @property
+    def kw_else(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ELSE)
+
+    @property
+    def kw_enum(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ENUM)
+
+    @property
+    def kw_const(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_CONST)
+
+    @property
+    def kw_pattern(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_PATTERN)
+
+    @property
+    def kw_properties(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_PROPERTIES)
+
+    @property
+    def kw_pattern_properties(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_PATTERN_PROPERTIES)
+
+    @property
+    def kw_additional_properties(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ADDITIONAL_PROPERTIES)
+
+    @property
+    def kw_min_length(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_MIN_LENGTH)
+
+    @property
+    def kw_max_length(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_MAX_LENGTH)
+
+    @property
+    def kw_items(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_ITEMS)
+
+    @property
+    def kw_min_items(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_MIN_ITEMS)
+
+    @property
+    def kw_max_items(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_MAX_ITEMS)
+
+    @property
+    def kw_unique_items(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_UNIQUE_ITEMS)
+
+    @property
+    def kw_contains(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_CONTAINS)
+
+    @property
+    def kw_required(self) -> Optional["SchemaNode"]:
+        return self.get_keyword(KW_REQUIRED)
+
+    @property
+    def title(self) -> Optional[str]:
+        title_kw = self.get_keyword(KW_TITLE)
+        if not title_kw:
+            return None
+        title = title_kw.literal
+        title = cast(str, title)
+        return title
+
+    @property
+    def property_name(self) -> Optional[str]:
+        return self.parent_key
+
+    @property
+    def property_display_name(self) -> Optional[str]:
+        """The name to display in documentation for this property.
+
+        This is simply the property name unless it is under "patternProperties" and it has a title,
+        in which case it is that title
+        """
+        if self.is_pattern_property:
+            return self.title or self.parent_key
+        if self.is_additional_properties:
+            return "Additional Properties"
+        return self.parent_key
+
     def should_be_a_link(self, config: GenerationConfiguration) -> bool:
         """Check if this node should be displayed as a link to another section of the schema in the context of
         the provided configuration.
@@ -409,14 +637,19 @@ def build_intermediate_representation(
 
         return loaded_schema
 
+    def _get_node_ref(schema: Union[int, str, List, Dict]) -> str:
+        if isinstance(schema, dict) and REF in schema:
+            return schema[REF]
+        return ""
+
     def _build_node(
         depth: int,
         html_id: str,
         schema_file_path: str,
         path_to_element: List[Union[str, int]],
         schema: Union[Dict, List, int, str],
-        is_property: bool = False,
-        is_pattern_properties: bool = False,
+        parent: Optional[SchemaNode] = None,
+        parent_key: Optional[str] = None,
     ) -> SchemaNode:
         """Recursively build a schema representation
 
@@ -426,13 +659,19 @@ def build_intermediate_representation(
         :param schema_file_path: Real path to the schema (absolute path with symlinks resolved)
         :param path_to_element: Path from the root of the schema to the current element
         :param schema: The JSON schema part being represented
-        :param is_property: The node is a property. Even if it is a required keyword, it should be treated as such
-        :param is_pattern_properties: True if this node is under the keyword "patternProperties"
         :return: A representation of the schema
         """
         schema_file_path = os.path.realpath(schema_file_path)
 
-        new_node = SchemaNode(depth, file=schema_file_path, path_to_element=path_to_element, html_id=html_id)
+        new_node = SchemaNode(
+            depth,
+            file=schema_file_path,
+            path_to_element=path_to_element,
+            html_id=html_id,
+            parent=parent,
+            parent_key=parent_key,
+            ref_path=_get_node_ref(schema),
+        )
         if html_id == "root":
             html_id = ""
 
@@ -441,6 +680,7 @@ def build_intermediate_representation(
         if isinstance(schema, dict):
             keywords = {}
             pattern_id = 1
+            is_property = parent_key in [KW_PROPERTIES, KW_PATTERN_PROPERTIES]
             for schema_key, schema_value in schema.items():
                 # These won't be needed to render the documentation.
                 # The definitions will be reached from references, otherwise they are useless
@@ -464,25 +704,23 @@ def build_intermediate_representation(
                 # Add the property name (correctly escaped) to the ID
                 new_html_id = html_id
                 new_depth = depth
-                key_is_property = schema_key in ["properties", "patternProperties"]
-                if not key_is_property:
+                if schema_key not in [KW_PROPERTIES, KW_PATTERN_PROPERTIES]:
                     new_depth += 1
                     new_html_id += "_" if html_id else ""
-                    if not is_pattern_properties:
+                    if not parent_key == KW_PATTERN_PROPERTIES:
                         new_html_id += escape_property_name_for_id(schema_key)
                     else:
                         new_html_id += f"pattern{pattern_id}"
                         pattern_id += 1
 
-                property_path = copy.deepcopy(path_to_element) + [schema_key]
                 keywords[schema_key] = _build_node(
                     new_depth,
                     new_html_id,
                     schema_file_path,
-                    property_path,
+                    copy.deepcopy(path_to_element) + [schema_key],
                     schema_value,
-                    is_property=key_is_property,
-                    is_pattern_properties=schema_key == "patternProperties",
+                    parent=new_node,
+                    parent_key=schema_key,
                 )
             new_node.keywords = keywords
         elif isinstance(schema, list):
@@ -492,7 +730,9 @@ def build_intermediate_representation(
                 new_html_id = html_id + ("_" if html_id else "") + "i" + str(i)
 
                 array_items.append(
-                    _build_node(depth + 1, new_html_id, schema_file_path, path_to_element + [i], element)
+                    _build_node(
+                        depth + 1, new_html_id, schema_file_path, path_to_element + [i], element, parent=new_node
+                    )
                 )
             new_node.array_items = array_items
 
@@ -544,7 +784,7 @@ def get_required_properties(schema_node: SchemaNode) -> List[str]:
 def get_undocumented_required_properties(schema_node: SchemaNode) -> List[str]:
     required_properties = get_required_properties(schema_node)
 
-    documented_properties = schema_node.keywords.get("properties")
+    documented_properties = schema_node.keywords.get(KW_PROPERTIES)
     documented_properties = documented_properties.keywords.keys() if documented_properties else []
 
     undocumented = []
@@ -799,7 +1039,7 @@ def generate_from_schema(
         link_to_reused_ref=link_to_reused_ref,
     )
 
-    template_folder = os.path.join(os.path.dirname(__file__), TEMPLATE_FOLDER)
+    template_folder = os.path.join(os.path.dirname(__file__), TEMPLATE_FOLDER, config.template_name)
     base_template_path = os.path.join(template_folder, TEMPLATE_FILE_NAME)
 
     md = markdown2.Markdown(extras={"fenced-code-blocks": {"cssclass": "highlight jumbotron"}, "tables": None})
@@ -878,7 +1118,7 @@ def generate_from_filename(
         config=config,
     )
 
-    copy_css_and_js_to_target(result_file_name, copy_css, copy_js)
+    copy_css_and_js_to_target(result_file_name, config)
 
     with open(result_file_name, "w", encoding="utf-8") as result_schema_doc:
         result_schema_doc.write(rendered_schema_doc)
@@ -911,29 +1151,32 @@ def generate_from_file_object(
 
     result = generate_from_schema(schema_file, config=config)
 
-    copy_css_and_js_to_target(result_file.name, config.copy_css, config.copy_js)
+    copy_css_and_js_to_target(result_file.name, config)
 
     result_file.write(result)
 
 
-def copy_css_and_js_to_target(result_file_path: str, copy_css: bool, copy_js: bool) -> None:
+def copy_css_and_js_to_target(result_file_path: str, config: GenerationConfiguration) -> None:
     """Copy the CSS and JS files needed to display the resulting page to the directory containing the result file"""
     files_to_copy = []
-    if copy_css:
+    if config.copy_css:
         files_to_copy.append(CSS_FILE_NAME)
-    if copy_js:
+    if config.copy_js:
         files_to_copy.append(JS_FILE_NAME)
     if not files_to_copy:
         return
 
     target_directory = os.path.dirname(result_file_path)
-    source_directory = os.path.dirname(__file__)
+    source_directory = os.path.join(os.path.dirname(__file__), TEMPLATE_FOLDER, config.template_name)
     if target_directory == source_directory:
         return
 
     for file_to_copy in files_to_copy:
+        source_file_path = os.path.join(source_directory, file_to_copy)
+        if not os.path.exists(source_file_path):
+            continue
         try:
-            shutil.copy(os.path.join(source_directory, file_to_copy), os.path.join(target_directory, file_to_copy))
+            shutil.copy(source_file_path, os.path.join(target_directory, file_to_copy))
         except shutil.SameFileError:
             print(f"Not copying {file_to_copy} to {os.path.abspath(target_directory)}, file already exists")
 
@@ -1013,7 +1256,10 @@ def _apply_config_cli_parameters(
     for parameter in config_cli_parameters:
         if "=" in parameter:
             parameter_name, parameter_value = parameter.split("=")
-            parameter_value = json.loads(parameter_value)
+            try:
+                parameter_value = json.loads(parameter_value)
+            except JSONDecodeError:
+                pass
         else:
             parameter_name = parameter
             if parameter_name.startswith("no_"):
