@@ -146,6 +146,12 @@ def test_additional_properties() -> None:
         ],
     )
 
+    explanation_nodes = soup.find_all(class_="additional-properties")
+    assert [node.text for node in explanation_nodes] == [
+        "Additional Properties of any type are allowed.",
+        "Each additional property must conform to the following schema",
+    ]
+
     # badge for the first sub type with additionalProperties: false
     badges = soup.find_all("span", class_=["no-additional"])
     assert len(badges) == 1
@@ -176,12 +182,7 @@ def test_array() -> None:
 
     tests.html_schema_doc_asserts.assert_property_names(soup, ["fruits", "vegetables", "veggieName", "veggieLike"])
     tests.html_schema_doc_asserts.assert_descriptions(
-        soup,
-        [
-            "A representation of a person, company, organization, or place",
-            "The name of the vegetable.",
-            "Do I like this vegetable?",
-        ],
+        soup, ["A schema with an array", "The name of the vegetable.", "Do I like this vegetable?",],
     )
     tests.html_schema_doc_asserts.assert_types(
         soup, ["object", "array of string", "string", "array", "object", "string", "boolean"]
@@ -229,7 +230,7 @@ def test_combining_one_of():
     """Test rendering of oneOf schema attribute in tabs"""
     soup = generate_case("combining_oneOf")
 
-    tests.html_schema_doc_asserts.assert_one_of_options(soup, 4)
+    tests.html_schema_doc_asserts.assert_one_of_options_names(soup, ["diskDevice", "diskUUID", "nfs", "tmpfs"])
     tests.html_schema_doc_asserts.assert_types(soup, ["object"] * 5)
     tests.html_schema_doc_asserts.assert_required(soup, [True])
 
@@ -244,6 +245,12 @@ def test_combining_not():
     assert definitions[0].text.lstrip().startswith("Must not be:")
 
 
+def test_combining_anyOf_option_names():
+    soup = generate_case("anyOf_option_names")
+
+    tests.html_schema_doc_asserts.assert_any_of_options_names(soup, ["diskDevice", "diskUUID", "Option 3", "tmpfs"])
+
+
 def test_with_default() -> None:
     """Test rendering of default values"""
     soup = generate_case("with_default")
@@ -450,3 +457,7 @@ def test_schema_with_keywords_as_properties() -> None:
     keywords = ["default", "examples", "if", "then", "else", "anyOf", "allOf", "not", "properties", "patternProperties"]
     tests.html_schema_doc_asserts.assert_property_names(soup, keywords)
     tests.html_schema_doc_asserts.assert_descriptions(soup, keywords)
+
+
+# TODO: test for uniqueItems
+# TODO: test for contains
